import { globalSetting, globalStyle } from '../global';

import { gl_Colors, gl_Shifts, gl_Shifts_for_bar, gl_Shifts_map, gl_WidthShifts, barShift, getRatio, screenHeightPixel } from './render-utils.js';
import { vertexShader, fragmentShader } from './render-shader.js';


// const { ChangoItem } = require('./types.d.ts');


class WebGL2WaveRender {
    /**
     * 
     * @param {{
     * grid: HTMLDivElement,
     * view: HTMLDivElement,
     * values: HTMLDivElement
     * }} elements
     * @param {{
     * chango: Record<string, {
     *      kind: string,
     *      wave: Array<number | string>,
     *      lineVao: WebGLVertexArrayObject
     *      maskVao: WebGLVertexArrayObject
     * }>
     * view: Array<{ ref: string }>,
     * currentWires: Set<{
     *      kind: string,
     *      link: string,
     *      name: string,
     *      size: number,
     *      parent: object,
     *      type: string
     * }>,
     * time: number
     * }} globalLookup
     * @param {{
     * width: number,
     * height: number,
     * xScale: number,
     * xOffset: number,
     * yOffset: number,
     * yStep: number,
     * yDuty: number
     * }} pstate 
     * @param { Array } plugins
     */
    constructor(elements, globalLookup, pstate, plugins) {
        const canvas = document.createElement('canvas');
        elements.view.replaceChildren(canvas);

        this.elements = elements;
        this.globalLookup = globalLookup;
        this.canvas = canvas;
        this.pstate = pstate;
        this.plugins = plugins;

        const gl = canvas.getContext('webgl2', {
            premultipliedAlpha: false,
            alpha: true,
            antialias: false,
            depth: false
        });
        this.webglLocation = this.initProgram(gl);
        const { lineVerticesMap, maskVerticesMap } = this.makeVertex();
        this.lineVerticesMap = lineVerticesMap;
        this.maskVerticesMap = maskVerticesMap;
        this.initData();
        this.animationHandler = undefined;
    }

    /**
     * 
     * @param {WebGL2RenderingContext} gl 
     * @returns {{
     * colors: WebGLUniformLocation,
     * shifts: WebGLUniformLocation,
     * scale: WebGLUniformLocation,
     * offset: WebGLUniformLocation,
     * pos: number,
     * widthShifts: WebGLUniformLocation,
     * gl: WebGL2RenderingContext
     * }} 
    */
    initProgram(gl) {
        const program = gl.createProgram();
        gl.attachShader(program, vertexShader.make(gl));
        gl.attachShader(program, fragmentShader.make(gl));
        gl.linkProgram(program);
        gl.useProgram(program);
    
        const webglLocation = {
            colors: gl.getUniformLocation(program, 'colors'),
            shifts:  gl.getUniformLocation(program, 'shifts'),
            scale:  gl.getUniformLocation(program, 'scale'),
            offset: gl.getUniformLocation(program, 'offset'),
            pos: gl.getAttribLocation(program, 'pos'),
            widthShifts: gl.getUniformLocation(program, 'widthShifts'),
            gl
        };
    
        return webglLocation;
    }

    /**
     * 
     * @returns {{
     *  lineVerticesMap: Map<string, Uint32Array>,
     *  maskVerticesMap: Map<string, Uint32Array>
     * }}
     */
    makeVertex() {
        const globalLookup = this.globalLookup;
        const time = globalLookup.time;
        const lineVerticesMap = new Map();
        const maskVerticesMap = new Map();
        if (globalSetting.prerender) {
            for (const id of Reflect.ownKeys(globalLookup.chango)) {
                const { lineVertices, maskVertices } = this.makeVertexByID(id);
                lineVerticesMap.set(id, lineVertices);
                maskVerticesMap.set(id, maskVertices);
            }
        }

        return { lineVerticesMap, maskVerticesMap };
    }

    /**
     * 
     * @param {string} id 
     * @returns {{
     *  lineVertices: Uint32Array
     *  maskVertices: Uint32Array
     * }}
     */
    makeVertexByID(id) {
        const globalLookup = this.globalLookup;
        const time = globalLookup.time;
        const signalItem = globalLookup.chango[id];

        const { kind, wave } = signalItem;
        if (kind === 'bit') {
            const { lineVertices, maskVertices } = this.makeBitVertex(wave, time);
            return { lineVertices, maskVertices };
        } else if (kind === 'vec') {
            const { lineVertices, maskVertices } = this.makeVecVertex(wave, time);
            return { lineVertices, maskVertices };
        } else if (kind === '') {
            const { lineVertices, maskVertices } = this.makeVecVertex(wave, time);
            return { lineVertices, maskVertices };        
        }
        return {
            lineVertices: undefined,
            maskVertices: undefined
        }
    }


    /**
     * @description 将wave 的值转化为渲染需要用到的值
     * @param {*} value 
     * @return {{
     *  y: number       // 裁剪空间的纵坐标
     *  color: number   // 颜色的索引 颜色rgb = gl_Colors[color]
     * }}
     */
    translateValue2RenderParameter(value) {
        switch (value) {
            case 0:             return { y: -1, color: 2 };       // 0 value
            case 1:             return { y: 1,  color: 3 };       // 1 value
            case 2: case 3:     return { y: -1, color: 4 };       // 不定态 x
            case 4: case 5:     return { y: 0,  color: 2 };       // 高阻态 z
            default:            return { y: -1, color: 7 };       // 其他，我也不知道还有啥
        }
    }

    /**
     * 
     * @param {{x: number, y: number, color: number} | undefined} p0 前一个点
     * @param {{x: number, y: number, color: number} | undefined} p1 当前的点
     * @param {{x: number, y: number, color: number} | undefined} p2 后一个点
     * @returns {number} 这是 widthshift 的索引，只需要 + 4 再 % 8 就能得到另一个 
     */
    makeWidthShiftIndexByPoints(p0, p1, p2) {
        if (p0 === undefined) {
            if (p1.y === p2.y) {
                return 0;
            } else if (p1.x === p2.x) {
                return 6;
            }
        } else if (p2 === undefined) {
            if (p1.y === p0.y) {
                return 0;
            } else if (p1.x === p0.x) {
                return 6;
            }
        } else {
            if (p0.x !== p1.x && p0.y === p1.y && p1.x === p2.x && p1.y !== p2.y) {
                if (p2.y > p1.y) {
                    return 1;
                } else {
                    return 7;
                }
            } else if (p0.x === p1.x && p0.y !== p1.y && p1.x !== p2.x && p1.y === p2.y) {
                if (p1.y > p0.y) {
                    return 1;
                } else {
                    return 7;
                }
            }
        }

        return 0;
    }

    /**
     * 
     * @param {number} x1 
     * @param {number} y1 
     * @param {number} x2 
     * @param {number} y2 
     * @param {number} color 
     * @param {number} wsIndex
     * @returns {number[]} 
     */
    makeRectangleVertices(x1, y1, x2, y2, color, wsIndex = 0) {
        const r1 = [x1, gl_Shifts_map.get(y2), color, wsIndex];
        const r2 = [x1, gl_Shifts_map.get(y1), color, wsIndex];
        const r3 = [x2, gl_Shifts_map.get(y1), color, wsIndex];
        const r4 = [x2, gl_Shifts_map.get(y2), color, wsIndex];
        return this.makeQuadVertices(r1, r2, r3, r4);
    }

    /**
     * 
     * @param {number[]} p1 
     * @param {number[]} p2 
     * @param {number[]} p3 
     * @param {number[]} p4 
     * @returns {number[]} 
     */
    makeQuadVertices(p1, p2, p3, p4) {
        return [
            ...p1, ...p2, ...p3,
            ...p1, ...p3, ...p4
        ]
    }


    /**
     * 
     * @param {Array<string | number>} wave 
     * @param { number } time 
     * @returns {{
     *  lineVertices: Uint32Array
     *  maskVertices: Uint32Array
     * }}
     */
    makeBitVertex(wave, time, debug = false) {
        const length = wave.length;
        // 先将节点数据转化为裁剪空间的坐标
        const perspectivePoints = [];

        for (let i = 0; i < length; ++ i) {
            // const currentWave = wave[(i === 0) ? 0 : (i - 1)];
            const currentWave = wave[i];
            const nextWave = (i === (length - 1)) ? wave[i] : wave[i + 1];

            const t1 = currentWave[0];
            const t2 = (i === (length - 1)) ? time : wave[i + 1][0];
            const value1 = currentWave[1];
            const value2 = nextWave[1];
            
            const renderParam1 = this.translateValue2RenderParameter(value1);
            const renderParam2 = this.translateValue2RenderParameter(value2);

            if (i === 0) {
                perspectivePoints.push({ x: t1, y: renderParam1.y, color: renderParam1.color });
                perspectivePoints.push({ x: t2, y: renderParam1.y, color: renderParam1.color });
            } else {
                const lastPoint = perspectivePoints.at(-1);
                if ((lastPoint.y !== renderParam1.y) || (lastPoint.color !== renderParam1.color)) {
                    perspectivePoints.push({ x: t1, y: renderParam1.y, color: renderParam1.color });
                }
                if ((renderParam1.y !== renderParam2.y) || (renderParam1.color !== renderParam2.color)) {
                    perspectivePoints.push({ x: t2, y: renderParam1.y, color: renderParam1.color });
                }
            }
        }

        // 确保最后一个点延申到了 time
        const lastPoint = perspectivePoints.at(-1);
        if (lastPoint.x < time) {
            perspectivePoints.push({ x: time, y: lastPoint.y, color: lastPoint.color });
        }

        // 计算出传入 shader 四元组数组
        // 四元组： (x, yshift_index, color_index, width_shift_index)
        const pointNum = perspectivePoints.length;
        const lineVertices = [];
        const maskVertices = [];

        // const lineVertices = [
        //     0, 0, 2, 0,
        //     10, 0, 2, 0,
        //     10, 4, 2, 0,
        //     20, 4, 2, 0,
        //     20, 0, 2, 0,
        //     30, 0, 2, 0,
        //     30, 4, 2, 0
        // ];
        // return new Uint32Array(lineVertices);
        

        // 制作 lineVertices
        for (let i = 0; i < pointNum; ++ i) {
            // p0: 上一个点
            // p1: 当前的点
            // p2: 下一个点
            const p0 = perspectivePoints[i - 1];
            const p1 = perspectivePoints[i];
            const p2 = perspectivePoints[i + 1];
            // p1.x, y1Index, p1.color, 0,
            const y1Index = gl_Shifts_map.get(p1.y);

            const wsIndex = this.makeWidthShiftIndexByPoints(p0, p1, p2);

            if (wsIndex === undefined) {
                console.log(p0, p1, p2);
            }

            lineVertices.push(
                p1.x, y1Index, p1.color, wsIndex,
                p1.x, y1Index, p1.color, (wsIndex + 4) % 8
            );

            // 防止颜色不同导致单个图元内出现两个颜色，这会引发shader的渐变
            if (p2 !== undefined && p1.color !== p2.color) {
                lineVertices.push(
                    p1.x, y1Index, p2.color, wsIndex,
                    p1.x, y1Index, p2.color, (wsIndex + 4) % 8
                );
            }
        }

        // 制作 maskVertices
        for (let i = 0; i < pointNum; ++ i) {
            const p1 = perspectivePoints[i];

            // 开头就有一条线
            if (i === 0 && p1.y === 1) {
                while (perspectivePoints[++ i] && perspectivePoints[i].y === 1);
                // 回退
                if (-- i > 0) {
                    // 四元组： (x, yshift_index, color_index, width_shift_index)
                    const rectangleVertices = this.makeRectangleVertices(p1.x, 1, perspectivePoints[i].x, -1, p1.color + 10, 4);
                    // 三角图元画矩形
                    maskVertices.push(...rectangleVertices);
                    continue;
                }
            }

            // 上升沿才使用
            const p2 = perspectivePoints[i + 1];
            const p3 = perspectivePoints[i + 2];

            if (p1 === undefined || p2 === undefined || p3 === undefined) {
                continue;
            }
            if (p2.y > p1.y) {
                // 矩形的四个点
                // 四元组： (x, yshift_index, color_index, width_shift_index)
                const rectangleVertices = this.makeRectangleVertices(p2.x, p2.y, p3.x, p1.y, p2.color + 10, 4);
                // 三角图元画矩形
                maskVertices.push(...rectangleVertices);
            }
        }

        if (debug) {
            console.log(perspectivePoints);
            console.log(pointNum);
            console.log(lineVertices);
        }

        return {
                lineVertices : new Uint32Array(lineVertices),
                maskVertices : new Uint32Array(maskVertices)
            };
    }




    /**
     * 
     * @param {Array<string | number>} wave 
     * @param { number } time 
     * @returns {{
     *  lineVertices: Uint32Array
     *  maskVertices: Uint32Array
     * }}
     */
    makeVecVertex(wave, time, debug = false) {
        const lineVertices = [];
        const maskVertices = [];
        const length = wave.length;

        for (let i = 0; i < length; ++ i) {
            const [t1, val, mask] = wave[i];
            const t2 = (i === (length - 1)) ? time : wave[i + 1][0];
            
            // t1(val) --- t2(val)
            // 详见 设计图 ./design/webgl.drawio makeVecVertex原理 sheet
            // 此处的 y 不是 y 的索引，详见 gl_Shifts_for_bar 的设计
            const p0 = {x: t1, y: 0};
            const p1 = {x: t2, y: 0};
            const a0 = {x: t1, y: 4};
            const a1 = {x: t2, y: 5};
            const a2 = {x: t1, y: 1};
            const a3 = {x: t2, y: 2};

            const color = mask ? 4 : 5;

            const points = [ a1, p1, a3, a2, p0, a0 ];
            const wsIndice = [ 1, 2, 3, 5, 6, 7 ];
            for (let i = 0; i < 6; ++ i) {
                const prePoint = points[i];
                const nextPoint = points[(i + 1) % 6];
                const preWsIndex = wsIndice[i];
                const nextWsIndex = wsIndice[(i + 1) % 6];
                const quadVertices = this.makeQuadVertices( 
                    [prePoint.x,  prePoint.y,  color, preWsIndex],
                    [prePoint.x,  prePoint.y,  color, (preWsIndex + 4) % 8],
                    [nextPoint.x, nextPoint.y, color, (nextWsIndex + 4) % 8],
                    [nextPoint.x, nextPoint.y, color, nextWsIndex],
                );
                lineVertices.push(...quadVertices);
            }

            const maskTriangles = [
                [3, 4, 5],
                [5, 0, 3],
                [0, 3, 2],
                [0, 1, 2]
            ];

            for (const triangleIndice of maskTriangles) {
                const [i1, i2, i3] = triangleIndice;
                const point1 = points[i1];
                const point2 = points[i2];
                const point3 = points[i3];
                maskVertices.push(
                    point1.x, point1.y, color + 10, wsIndice[i1],
                    point2.x, point2.y, color + 10, wsIndice[i2],
                    point3.x, point3.y, color + 10, wsIndice[i3]
                );
            }
        }

        if (debug) {
            console.log(lineVertices);
        }

        return {
            lineVertices: new Uint32Array(lineVertices),
            maskVertices: new Uint32Array(maskVertices)
        };
    }

    /**
     * 
     * @param {string} id 
     * @returns 
     */
    initVertice(id) {
        const webglLocation = this.webglLocation;
        const gl = webglLocation.gl;
        const signalItem = this.globalLookup.chango[id];

        if (this.lineVerticesMap.get(id) === undefined) {
            const { lineVertices, maskVertices } = this.makeVertexByID(id);
            this.lineVerticesMap.set(id, lineVertices);
            this.maskVerticesMap.set(id, maskVertices);
        }

        const lineVertices = this.lineVerticesMap.get(id);
        const maskVertices = this.maskVerticesMap.get(id);

        // 创建并设置 绘制wave轮廓 主体轮廓的 缓冲区、vao、顶点设置
        const lineVertexBuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, lineVertexBuffer);
        gl.bufferData(gl.ARRAY_BUFFER, lineVertices, gl.STATIC_DRAW);
        signalItem.lineVao = gl.createVertexArray();
        gl.bindVertexArray(signalItem.lineVao);
        gl.vertexAttribIPointer(webglLocation.pos, 4, gl.UNSIGNED_INT, 0, 0);
        gl.enableVertexAttribArray(webglLocation.pos);

        // 创建并设置 绘制wave半透明遮罩层 主体轮廓的 缓冲区、vao、顶点设置
        const maskVertexBuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, maskVertexBuffer);
        gl.bufferData(gl.ARRAY_BUFFER, maskVertices, gl.STATIC_DRAW);
        signalItem.maskVao = gl.createVertexArray();
        gl.bindVertexArray(signalItem.maskVao);
        gl.vertexAttribIPointer(webglLocation.pos, 4, gl.UNSIGNED_INT, 0, 0);
        gl.enableVertexAttribArray(webglLocation.pos);
    }

    initData() {
        if (globalSetting.prerender) {
            for (const id of Reflect.ownKeys(this.globalLookup.chango)) {
                this.initVertice(id);
            }
        } 

        const webglLocation = this.webglLocation;
        const gl = webglLocation.gl;
        gl.uniform4fv(webglLocation.colors, gl_Colors);
        gl.uniform2fv(webglLocation.widthShifts, gl_WidthShifts);
        gl.uniform2fv(webglLocation.shifts, gl_Shifts);
    }

    /**
     * 
    * @param {Array<{
    *  index: number
    *  rgba: {
    *   red: number,
    *   green: number,
    *   blue: number,
    *   alpha: number
    * }
    * }>} updaters 
    * @param {{
    *  updateMask: boolean
    * }} config 
     */
    updateGLColor(updaters, config) {
        const webglLocation = this.webglLocation;
        const gl = webglLocation.gl;

        for (const updater of updaters) {
            const startIndex = updater.index * 4;
            const rgba = updater.rgba;
            gl_Colors[startIndex]     = rgba.red;
            gl_Colors[startIndex + 1] = rgba.green;
            gl_Colors[startIndex + 2] = rgba.blue;
            gl_Colors[startIndex + 3] = rgba.alpha;

            if (config.updateMask) {
                const maskIndex = (updater.index + 10) * 4;
                gl_Colors[maskIndex]     = rgba.red;
                gl_Colors[maskIndex + 1] = rgba.green;
                gl_Colors[maskIndex + 2] = rgba.blue;
            }
        }

        gl.uniform4fv(webglLocation.colors, gl_Colors);
        this.render();
    }

    /**
     * 
     * @param {{
     *  type: 'common' | 'action'
     * } | undefined} renderConfig 
     */
    render(renderConfig) {
        renderConfig = renderConfig || { type: 'common' };
        const needHiddenAnimation = globalSetting.renderAnimation && renderConfig.type === 'action';

        if (this.animationHandler !== undefined) {
            cancelAnimationFrame(this.animationHandler);
        }
        
        const canvas = this.canvas;
        const webglLocation = this.webglLocation;
        const gl = webglLocation.gl;
        const globalLookup = this.globalLookup;
        const lineVerticesMap = this.lineVerticesMap;
        const maskVerticesMap = this.maskVerticesMap;
        const elements = this.elements;

        // 更新 yDuty & yStep
        // yStep: scale 之后的单个 wave 的容器高度需要和 displaySignalHeight + sideBarItemMargin 一样
        // yDuty: scale 之后的单个 wave 的 padding 必须等于 sideBarItemMargin，也就是 (1 - yDuty) * yStep = 2 * margin
        const sidebarItemMargin = globalStyle.sideBarItemMargin;

        const pstate = this.pstate;
        const plugins = this.plugins;
        
        pstate.yStep = globalSetting.displaySignalHeight + globalStyle.sideBarItemMargin;
        pstate.yDuty = (1 - 2 * globalStyle.sideBarItemMargin / this.pstate.yStep) * 0.9;
        
        document.body.style.setProperty('--vcd-render-padding',  pstate.topBarHeight + 'px');
        const canvasHeight = pstate.height - pstate.topBarHeight - pstate.botBarHeight;
        const canvasWidth = pstate.width;

        // 默认 1594
        canvas.width = canvasWidth;
        // 默认 1260
        canvas.height = canvasHeight;

        let startTime = undefined;
        // 默认执行一个 300 ms 的动画
        const animationDuration = 300;
        
        if (needHiddenAnimation) {
            this.elements.grid.classList.add('vcd-hidden');
            this.elements.values.classList.add('vcd-hidden');
        }
        
        let animationHandler = this.animationHandler = requestAnimationFrame(drawWaves);

        const _this = this;

        function linearAnimation(delta, oldVal, newVal) {
            return (1 - delta) * oldVal + delta * newVal;
        }

        function cubicBezierAnimation(delta, oldVal, newVal) {
            delta = 3 * (1 - delta) * (1 - delta) * delta + 3 * (1 - delta) * delta * delta + delta * delta * delta;
            return (1 - delta) * oldVal + delta * newVal;
        }

        function renderOneFrame(animationDelta) {
            // 考虑设置线性动画的变量 xScale, xOffset, yOffset, yStep, yDuty
            const xScale  = cubicBezierAnimation(animationDelta, pstate.oldXScale, pstate.xScale);
            const xOffset = cubicBezierAnimation(animationDelta, pstate.oldXOffset, pstate.xOffset);
            const yOffset = cubicBezierAnimation(animationDelta, pstate.oldYOffset, pstate.yOffset);
            const yStep   = cubicBezierAnimation(animationDelta, pstate.oldYStep, pstate.yStep);
            const yDuty   = cubicBezierAnimation(animationDelta, pstate.oldYDuty, pstate.yDuty);

            // 设置 glsl 变量
            const scaleX = 2 * xScale / canvasWidth;
            const scaleY = yStep * yDuty / canvasHeight;
            gl.uniform2f(webglLocation.scale, scaleX, scaleY);

            // 设置 webgl 和 canvas 大小位置一致
            gl.viewport(0, 0, canvasWidth, canvasHeight);
            // 清楚颜色缓冲区，也就是删除上一次的渲染结果
            gl.clear(gl.COLOR_BUFFER_BIT);

            // 根据 globalLookup 当前激活的需要渲染的信号进行渲染
            let index = 0;
            for (const signal of globalLookup.currentWires) {
                const id = signal.link;
                const wave = globalLookup.chango[id].wave;

                // _this.makeBitVertex(wave, globalLookup.time, true);
                // _this.makeVecVertex(wave, globalLookup.time, true);

                const signalItem = globalLookup.chango[id];
                if (!signalItem) {
                    return;
                }

                const offsetX = 2 * xOffset / canvasWidth - 1;
                const offsetY = 2 * (yOffset - yStep * index) / canvasHeight + 1;
                gl.uniform2f(webglLocation.offset, offsetX, offsetY);

                // 根据 lineVao 进行绘制            
                if (lineVerticesMap.get(id) === undefined) {
                    _this.initVertice(id);
                }
                const lineVertices = lineVerticesMap.get(id);
                const maskVertices = maskVerticesMap.get(id);

                if (signal.size === 1) {
                    // 如果是 width 为 1 的
                    gl.uniform2fv(webglLocation.shifts, gl_Shifts);
                    gl.bindVertexArray(signalItem.lineVao);
                    gl.drawArrays(gl.TRIANGLE_STRIP, 0, lineVertices.length / 4);

                    gl.bindVertexArray(signalItem.maskVao);
                    gl.drawArrays(gl.TRIANGLES, 0, maskVertices.length / 4);
                } else {
                    // 如果是 width 大于 1 的
                    gl.uniform2fv(webglLocation.shifts, gl_Shifts_for_bar);
                    gl.bindVertexArray(signalItem.lineVao);
                    gl.drawArrays(gl.TRIANGLES, 0, lineVertices.length / 4);

                    gl.bindVertexArray(signalItem.maskVao);
                    gl.drawArrays(gl.TRIANGLES, 0, maskVertices.length / 4);
                }

                index ++;
            }

            plugins.map(fn => fn(globalLookup, pstate, elements));
        }

        function drawWaves(now) {
            if (startTime === undefined) {
                startTime = now;
            } else {
                const pastTime = now - startTime;
                const animationDelta = globalSetting.renderAnimation ? Math.min(pastTime / animationDuration, 1) : 1;
                renderOneFrame(animationDelta);
                // 达到 1 时说明线性动画时间到了，退出即可
                if (animationDelta >= 1) {
                    cancelAnimationFrame(animationHandler);
                    // 更新状态变量
                    pstate.oldXOffset = pstate.xOffset;
                    pstate.oldYOffset = pstate.yOffset;
                    pstate.oldXScale = pstate.xScale;
                    pstate.oldYStep = pstate.yStep;
                    pstate.oldYDuty = pstate.yDuty;

                    if (needHiddenAnimation) {
                        _this.elements.grid.classList.remove('vcd-hidden');
                        _this.elements.values.classList.remove('vcd-hidden');
                    }
                    return;
                }
            }
            // 进入下一帧
            requestAnimationFrame(drawWaves);
        }
    }

}

export default WebGL2WaveRender;